/*  Copyright (C) 2013 Gerald Mwangi

	This file is part of Charon.

	Charon is free software: you can redistribute it and/or modify
	it under the terms of the GNU Lesser General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	Charon is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU Lesser General Public License for more details.

	You should have received a copy of the GNU Lesser General Public License
	along with Charon.  If not, see <http://www.gnu.org/licenses/>.
*/
/** \file LinearMotionCompensation.hxx
 *  Implementation of the parameter class LinearMotionCompensation.
 *  \author <a href="mailto:gerald.mwangi@gmx.de">
 *      Gerald Mwangi</a>
 *  \date 19.06.2013
 */

#ifndef _LINEARMOTIONCOMPENSATION_HXX_
#define _LINEARMOTIONCOMPENSATION_HXX_

#include <charon-core/ParameteredObject.hxx>
#include "LinearMotionCompensation.h"
#include <opencv2/nonfree/nonfree.hpp>
using namespace vigra;
using namespace std;
using namespace cv;
template <typename T>
LinearMotionCompensation<T>::LinearMotionCompensation(const std::string& name) :
		TemplatedParameteredObject<T>(
			"LinearMotionCompensation", name,
			"<h2>Compensates linear motion in an image sequence</h2><br>"
			"Compensates linear motion in an image sequence, using the sift "
			"detector in the opencv library. The motion is assumed to be "
			"constrained to the image plane (x,y coordinates), thus no "
			"motion in z direction."
		)
{
	ParameteredObject::_addInputSlot(
		image_sequence, "image_sequence",
		"sequence with linear motion",
		"vigraArray5<T>");
	ParameteredObject::_addInputSlot(
		roi, "roi",
		"roi to estimate the linear motion",
		"Roi<T>*");

	ParameteredObject::_addOutputSlot(
		compensated_sequence, "compensated_sequence",
		"motion compensated sequence",
		"vigraArray5<T>");
	ParameteredObject::_addOutputSlot(
		feature_sequence, "feature_sequence",
		"image sequence with found features",
		"vigraArray5<T>");

	ParameteredObject::_addParameter< double >(
		motion_tolerance, "motion_tolerance",
		"allowed deviation from linear motion model",
		0.5, "double");
	ParameteredObject::_addParameter< double >(
		good_feature_fraction, "good_feature_fraction",
		"Fraction of the number features with motion tolerance below"
		"motion_tolerance parameter and total number of found features."
		"Values ~1 mean nearly all features are required to have a linear"
		"motion, values ~0 mean no features must have a linear motion.",
		0.5, "double");

}

template <typename T>
void LinearMotionCompensation<T>::execute() {
    Mat src,dst,dstkpts;
    const vigra::MultiArrayView<5, T>& i=image_sequence();
    if(i.size(2)<2)
        ParameteredObject::raise("Insufficient number of images in input sequence");

    typename vigra::MultiArrayShape<5>::type feature_shape=image_sequence().shape();
    feature_shape[4]=3;
    feature_sequence().reshape(feature_shape);
    compensated_sequence()=image_sequence();
    getCvMatFrame(src,0,true);
    getCvMatFrame(dst,1,true);
    dst=warpImage(src,dst,dstkpts);
    copyFrameToOut(dst,dstkpts,1);
    //namedWindow("matches", 1);
    for(int ii=2;ii<i.size(2);ii++)
    {
        getCvMatFrame(dst,ii,false);
        dst=warpImage(src,dst,dstkpts);
        copyFrameToOut(dst,dstkpts,ii);
    }





}

template <typename T>
void LinearMotionCompensation<T>::getCvMatFrame(cv::Mat &out, int frame,bool create)
{
    const vigra::MultiArrayView<5, T>& i=image_sequence();
    int width=i.size(0);
    int height=i.size(1);
    if(create)
    {
        string type=this->getTemplateType();
        if(type=="double")
            out.create(width,height,CV_64F);
        else if(type=="float")
            out.create(width,height,CV_32F);
        else
            out.create(width,height,CV_32S);
    }
    for(int x=0;x<width;x++)
        for(int y=0;y<height;y++)
            out.at<T>(x,y)=i(x,y,frame);
}

template <typename T>
cv::Mat LinearMotionCompensation<T>::warpImage(cv::Mat src, cv::Mat dst,cv::Mat& dst_imkpts)
{
    cv::Mat src16;
    src.copyTo(src16);
    cv::Mat dst16;
    dst.copyTo(dst16);
    cv::SiftFeatureDetector detector(0,3,0.001);

    cv::SiftDescriptorExtractor descrExt;
    vector<cv::KeyPoint> dstKpts,srcKpts;
    double maxsrc=0,maxdst=0;
    double minsrc=0,mindst=0;
    int dummy=0;
    cv::minMaxIdx(src,&minsrc,&maxsrc);
    cv::minMaxIdx(dst,&mindst,&maxdst);
    cout<<"src max "<<maxsrc<<" min "<<minsrc<<endl;
    cout<<"dst max "<<maxdst<<" min "<<mindst<<endl;

    src.convertTo(src,CV_8U,255/(maxsrc-minsrc),-255*minsrc/(maxsrc-minsrc));
    dst.convertTo(dst,CV_8U,255/(maxdst-mindst),-255*mindst/(maxdst-mindst));
//    cv::imshow("test",dst);
//    waitKey(0);
    cv::BFMatcher matcher(cv::NORM_L1);

    detector.detect(dst,dstKpts);
    detector.detect(src,srcKpts);

    //srcKpts=keypointsInRoi(reduceKeypointsByNeigSize(srcKpts,20),100,200,-1,-1);
    //dstKpts=keypointsInRoi(reduceKeypointsByNeigSize(dstKpts,20),100,200,-1,-1);

    cv::Mat srcDescr,dstDescr;
    descrExt.compute(src,srcKpts,srcDescr);
    descrExt.compute(dst,dstKpts,dstDescr);

    vector< cv::DMatch > match,maskedmatch;
    matcher.match(srcDescr,dstDescr,match);


    vector<cv::KeyPoint> matched_dstkpts;
    for(int i=0;i<match.size();i++)
    {
        cv::Point2f srcpt=srcKpts[match[i].queryIdx].pt;
        if(roi.connected())
        {
            if(roi()->xBegin!=roi()->xEnd&&roi()->yBegin!=roi()->yEnd)
            {
                if(srcpt.x>roi()->xBegin&&srcpt.x<roi()->xEnd&&srcpt.y>roi()->yBegin&&srcpt.y<roi()->yEnd)
                {
                    maskedmatch.push_back(match[i]);
                    matched_dstkpts.push_back(dstKpts[((cv::DMatch)match[i]).trainIdx]);
                }
            }
            else if(roi()->xBegin!=roi()->xEnd)
            {
                if(srcpt.x>roi()->xBegin&&srcpt.x<roi()->xEnd)
                {
                    maskedmatch.push_back(match[i]);
                    matched_dstkpts.push_back(dstKpts[((cv::DMatch)match[i]).trainIdx]);
                }
            }
            else if(roi()->yBegin!=roi()->yEnd)
            {
                if(srcpt.y>roi()->yBegin&&srcpt.y<roi()->yEnd)
                {
                    maskedmatch.push_back(match[i]);
                    matched_dstkpts.push_back(dstKpts[((cv::DMatch)match[i]).trainIdx]);
                }
            }
        }
        else
        {
            maskedmatch.push_back(match[i]);
            matched_dstkpts.push_back(dstKpts[((cv::DMatch)match[i]).trainIdx]);
        }

    }
    cv::Mat affTrans,res;
    src16.copyTo(res);
    cv::drawKeypoints(dst,matched_dstkpts,dst_imkpts);
//        Mat img_matches;
//        drawMatches(src, srcKpts, dst, dstKpts, maskedmatch, img_matches);
//        imshow("matches", img_matches);
//        waitKey(0);
    findAffineTrans(maskedmatch,srcKpts,dstKpts,motion_tolerance(),good_feature_fraction(),500,affTrans);






    //cv::warpPerspective(dst,res, H,cv::Size(src.cols,src.rows),CV_WARP_INVERSE_MAP);

    cv::warpAffine(dst16,res, affTrans,cv::Size(src.cols,src.rows),cv::INTER_LINEAR,cv::BORDER_TRANSPARENT);
    return res;

}

template <typename T>
void LinearMotionCompensation<T>::determinParameters(vector< cv::Point2f > best_src,vector< cv::Point2f > best_dst,float& cos_theta,cv::Point2f& trans)
{
    float cos_nom=0,cos_denom=0;
    cos_theta=0;
    cv::Point2f meansrc(0,0),meandst(0,0);
    for(int i=0;i<best_src.size();i++)
    {
    cv::Point2f srcpt=best_src[i];
    cv::Point2f dstpt=best_dst[i];
    meansrc=meansrc+srcpt;
    meandst=meandst+dstpt;

    }
    meansrc.x=meansrc.x/best_src.size();
    meansrc.y=meansrc.y/best_src.size();
    meandst.x=meandst.x/best_src.size();
    meandst.y=meandst.y/best_src.size();
    for(int i=0;i<best_src.size();i++)
    {
    cv::Point2f srcpt=best_src[i]-meansrc;
    cv::Point2f dstpt=best_dst[i]-meandst;
    cos_nom+=srcpt.dot(dstpt);
    cos_denom+=sqrt(srcpt.dot(srcpt)*dstpt.dot(dstpt));
    //cos_theta+=srcpt.dot(dstpt)/sqrt(srcpt.dot(srcpt)*dstpt.dot(dstpt));
    }
    //cos_theta/=best_src.size();
    //cos_theta=meansrc.dot(meandst)/(sqrt(meansrc.dot(meansrc)*meandst.dot(meandst)));//1;//cos_nom/cos_denom;
    cos_theta=cos_nom/cos_denom;
    float sin_theta=std::sqrt(1-cos_theta*cos_theta);

    trans.x=meandst.x-(cos_theta*meansrc.x-sin_theta*meansrc.y);
    trans.y=meandst.y-(sin_theta*meansrc.x+cos_theta*meansrc.y);
}

template <typename T>
float LinearMotionCompensation<T>::errorFunction(float cos_theta,cv::Point2f trans,cv::Point2f src,cv::Point2f dst)
{
    cv::Point2f errpt;

    float sin_theta=std::sqrt(1-cos_theta*cos_theta);
    errpt.x=cos_theta*src.x-sin_theta*src.y+trans.x-dst.x;
    errpt.y=sin_theta*src.x+cos_theta*src.y+trans.y-dst.y;
    float res=errpt.dot(errpt)/2;
    //float res=sqrt(errpt.dot(errpt));
    return res;
}


template <typename T>
void LinearMotionCompensation<T>::findAffineTrans(vector< cv::DMatch >  matches,vector<cv::KeyPoint> srckpts,vector<cv::KeyPoint> dstkpts,float tolerance,float inlier_frac,int max_iter,cv::Mat& affTrans)
{
    float cos_theta=1;
    cv::Point2f trans;
    vector< cv::Point2f > best_consensus_src,best_consensus_dst;
    for(int iter=0;iter<max_iter;iter++)
    {
    //Get min sample set
    int sample1=double(rand())/(RAND_MAX)*(matches.size()-1);
    int sample2=double(rand())/(RAND_MAX)*(matches.size()-1);
    while(sample2==sample1)
        sample2=double(rand())/(RAND_MAX)*(matches.size()-1);
    cv::Point2f srcpt1=srckpts[matches[sample1].queryIdx].pt;
    cv::Point2f dstpt1=dstkpts[matches[sample1].trainIdx].pt;

    cv::Point2f srcpt2=srckpts[matches[sample2].queryIdx].pt;
    cv::Point2f dstpt2=dstkpts[matches[sample2].trainIdx].pt;

    //Solve for parameters
    cv::Point2f meansrc,meandst;
    meansrc.x=(srcpt1.x+srcpt2.x)/2;
    meansrc.y=(srcpt1.y+srcpt2.y)/2;

    meandst.x=(dstpt1.x+dstpt2.x)/2;
    meandst.y=(dstpt1.y+dstpt2.y)/2;

    srcpt1=srcpt1-meansrc;
    srcpt2=srcpt2-meansrc;

    dstpt1=dstpt1-meandst;
    dstpt2=dstpt2-meandst;

    //cos_theta=(srcpt1.dot(dstpt1)+srcpt2.dot(dstpt2))/(srcpt1.dot(srcpt1)+srcpt2.dot(srcpt2));

    float sin_theta=std::sqrt(1-cos_theta*cos_theta);

    trans.x=meandst.x-(cos_theta*meansrc.x-sin_theta*meansrc.y);
    trans.y=meandst.y-(sin_theta*meansrc.x+cos_theta*meansrc.y);


    //Determin consensus set
    vector< cv::Point2f > consensus_set_src,consensus_set_dst;
    for(int m=0;m<matches.size();m++)
    {
        cv::Point2f srccand=srckpts[matches[m].queryIdx].pt;
        cv::Point2f dstcand=dstkpts[matches[m].trainIdx].pt;
        float errcand=errorFunction(cos_theta,trans,srccand,dstcand);
        if(errcand<=tolerance)
        {
        consensus_set_src.push_back(srccand);
        consensus_set_dst.push_back(dstcand);
        }

    }

    //Is consensus set appropriate?
    float cons_frac=((float)consensus_set_src.size())/((float)matches.size());
    if(cons_frac>=inlier_frac)
    {
        best_consensus_src=consensus_set_src;
        best_consensus_dst=consensus_set_dst;
        break;
    }


    }
    determinParameters(best_consensus_src,best_consensus_dst,cos_theta,trans);
    affTrans.create(2,3,CV_32F);
    affTrans.at<float>(0,0)=cos_theta;
    affTrans.at<float>(1,1)=cos_theta;
    float sin_theta=std::sqrt(1-cos_theta*cos_theta);
    affTrans.at<float>(0,1)=sin_theta;
    affTrans.at<float>(1,0)=-sin_theta;
    affTrans.at<float>(0,2)=-trans.x;
    affTrans.at<float>(1,2)=-trans.y;

}

template <typename T>
void LinearMotionCompensation<T>::copyFrameToOut(Mat &in,Mat &inkpts, int frame)
{
    vigra::MultiArrayView<5, T>& out=compensated_sequence();
    vigra::MultiArrayView<5, T>& outkpts=feature_sequence();
    for(int x=0;x<out.size(0);x++)
        for(int y=0;y<out.size(1);y++)
        {
            out(x,y,frame)=in.at<T>(x,y);
            cv::Vec<uchar,3> color=inkpts.at<cv::Vec<uchar,3> >(x,y);
            outkpts(x,y,frame,0,0)=color.val[0];
            outkpts(x,y,frame,0,1)=color.val[1];
            outkpts(x,y,frame,0,2)=color.val[2];

        }
}


#endif /* _LINEARMOTIONCOMPENSATION_HXX_ */
